"""Optimize the Enumeration of an encoding using the mutual information."""

import logging
from functools import partial

import numpy as np
from numpy.typing import NDArray

from .enumeration.evolutionary import lambda_plus_mu

logger = logging.getLogger(__name__)


def distance_squared(
    mutual_information: NDArray, permutation: list | NDArray
) -> list[float]:
    """Cost function weighting by the mutualinformation and mode index difference squared.

    Args:
        mutual_information (NDArray): The mutual information of modes.
        permutation (list | NDArray): A permutation of mode labels from 0 to M.

    Returns:
        list[float]: Calculated cost as a length 1 list. (Needed for DEAP.)
    distance matrix(6) = array([
        [ 0,  1,  4,  9, 16, 25],
        [ 1,  0,  1,  4,  9, 16],
        [ 4,  1,  0,  1,  4,  9],
        [ 9,  4,  1,  0,  1,  4],
        [16,  9,  4,  1,  0,  1],
        [25, 16,  9,  4,  1,  0]
        ])

    Example:
        >>> import numpy as np
        >>> from ferrmion.optimize.enumeration.cost_functions import distance_squared
        >>> mi = np.ones((3,3))
        >>> distance_squared(mi, [0,1,2])
    """
    n_mode = mutual_information.shape[0]
    if set(permutation) != set(range(n_mode)):
        logger.warning("Invalid permutation %s, returning infinite cost.", permutation)
        return [np.inf]
    distance_matrix = np.array(
        [[np.abs(i - j) ** 2 for j in range(n_mode)] for i in range(n_mode)]
    )
    return [
        np.sum(mutual_information[permutation, :][:, permutation] * distance_matrix)
    ]


def minimise_mi_distance(
    mutual_information: NDArray,
    pop_size: int = 500,
    ngen: int = 50,
    pair_spins: bool = False,
    spinless_mi: bool = True,
) -> NDArray:
    """Place modes with high mutual information near eachother.

    This uses the lambda_plus_mu evolutionary algorithm.

    Args:
        mutual_information (NDArray): A 2d array of mode mutual information values.
        pop_size (int): The size of the initial population.
        ngen (int): The number of generations to evolve.
        pair_spins (bool): Pair the alpha and beta spins so that they remain adjacent in the mode ordering.
        spinless_mi (bool): Whether the given mutual information matrix has seperate spin orbitals.

    Returns:
        NDArray: The best mode ordering found.

    Example:
        >>> import numpy as np
        >>> from ferrmion.optimize.enumeration.cost_functions import minimise_mi_distance
        >>> mi = 0.5 * np.random.random((3,3))
        >>> mi = mi + mi.T
        >>> minimise_mi_distance(mi, pop_size=10, ngen=2)
    """
    logger.debug("Minimising disance between high MI modes.")
    if not spinless_mi and pair_spins:
        logger.debug("Pairing spins.")
        squash_rows = mutual_information[::2] + mutual_information[1::2]
        mutual_information = squash_rows[:, ::2] + squash_rows[:, 1::2]

    evaluate = partial(distance_squared, mutual_information)
    n_modes: int = mutual_information.shape[0]
    best, _ = lambda_plus_mu(n_modes, evaluate, pop_size, ngen)

    if pair_spins:
        best = np.stack((2 * best, (2 * best) + 1)).T.flatten()
    logger.debug("Found best ordering %s", best)
    return best


def coefficient_pauli_weight(pauli_hamiltonian: dict[str, float]) -> list[float]:
    """The Pauli-weight of a template scaled by the term coefficients.

    Args:
        pauli_hamiltonian (dict[bytes, float]): A filled template hamiltonian with byte-hashed keys.

    Return:
        list[float]: A single value in a list (needed for deap) giving the cost.

    Example:
        >>> from ferrmion.optimize.enumeration.cost_functions import coefficient_pauli_weight
        >>> from ferrmion.utils import symplectic_hash
        >>> hashed_vec = symplectic_hash(np.array([True, False, False, True]))
        >>> coefficient_pauli_weight({hashed_vec:1}, [0,1,2])
    """
    logger.debug("Calculating Pauli-weighted Norm")
    logger.debug(pauli_hamiltonian)

    weighted_terms = [
        (len(k) - k.count("I")) * np.abs(v) for k, v in pauli_hamiltonian.items()
    ]
    logger.debug(weighted_terms)
    return [np.sum(weighted_terms)]


def pauli_weight(pauli_hamiltonian: dict[str, float]) -> list[float]:
    """The Pauli-weight of a template scaled by the term coefficients.

    Args:
        pauli_hamiltonian (dict[bytes, float]): A filled template hamiltonian with byte-hashed keys.

    Return:
        list[float]: A single value in a list (needed for deap) giving the cost.

    Example:
        >>> from ferrmion.optimize.enumeration.cost_functions import coefficient_pauli_weight
        >>> from ferrmion.utils import symplectic_hash
        >>> hashed_vec = symplectic_hash(np.array([True, False, False, True]))
        >>> coefficient_pauli_weight({hashed_vec:1}, [0,1,2])
    """
    logger.debug("Calculating Pauli-weighted Norm")
    logger.debug(pauli_hamiltonian)

    weighted_terms = [(len(k) - k.count("I")) for k, v in pauli_hamiltonian.items()]
    logger.debug(weighted_terms)
    return [np.sum(weighted_terms)]



"""Ternary Tree fermion to qubit mappings."""
from ferrmion.hamiltonians import FermionHamiltonian, QubitHamiltonian
from ferrmion.core import encode, topphatt

import logging

import numpy as np
from numpy.typing import NDArray

from .base import FermionQubitEncoding
from ferrmion.core import encode_standard
from .ternary_tree_node import TTNode, node_sorter

logger = logging.getLogger(__name__)


class TernaryTree(FermionQubitEncoding):
    """Ternary tree encoding for fermionic operators.

    Attributes:
        n_modes (int): The number of fermionic modes to be encoded.
        n_qubits (int): The number of qubits in encoded operators.
        root (TTNode): The root node of the tree.
        enumeration_scheme (dict[str, tuple[int, int]] | None): The enumeration scheme.

    Methods:
        default_mode_op_map(): Create a default mode operator map for the tree.
        default_enumeration_scheme(): Create a default enumeration scheme for the tree.
        as_dict(): Return the tree structure as a dictionary.
        add_node(node_string: str): Add a node to the tree.
        branch_pauli_map(): Create a map from each branch string to a Pauli string.
        string_pairs(): Return the pair of branch strings which correspond to each node.
        _build_symplectic_matrix(): Build the symplectic matrix for the tree.

    Simple Example:
        >>> from ferrmion.encode.ternary_tree import TernaryTree
        >>> tree = TernaryTree(4)
        >>> tree.add_node('x')
        >>> tree.enumeration_scheme = tree.default_enumeration_scheme()
        >>> tree.as_dict()

    Advanced Usage:
        >>> from ferrmion.encode.ternary_tree import TernaryTree
        >>> tree = TernaryTree(4)
        >>> jordan_wigner = tree.JW()
        >>> bravyi_kitaev = tree.BK()
        >>> parity = tree.Parity()
        >>> minimum_height_tree = tree.JKMN()
    """

    def __init__(
        self,
        n_modes: int,
        n_qubits: None | int = None,
        root_node: TTNode = TTNode(),
    ):
        """Initialise a ternary tree.

        Args:
            n_modes (int): How many fermionic modes in the encoding.
            n_qubits (int): Optional overwrite of number of qubits in target encoding.
            root_node (TTNode): The root node of the tree.
        """
        self.n_modes = n_modes
        self.n_qubits = n_modes if n_qubits is None else n_qubits
        self.root_node = root_node

        if None not in root_node.child_qubit_labels.values():
            self.enumeration_scheme: dict[str, tuple[int, int]] = {
                node: (mode, qubit)
                for mode, (node, qubit) in enumerate(
                    root_node.child_qubit_labels.items()
                )
            }

        self.vacuum_state = np.array([0] * self.n_qubits, dtype=np.uint8)
        # self._enumeration_scheme = {}
        super().__init__(self.n_modes, self.n_qubits)

    @classmethod
    def from_hamiltonian_coefficients(cls, coeffs: tuple) -> "TernaryTree":
        """Create an encoding by passing coefficients.

        Args:
            coeffs (tuple): The electron integrals for some hamiltonian.

        Returns:
            FermionQubitEncoding: An initialised encoding.

        Example:
            >>> import numpy as np
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> coeffs = (np.zeros((4, 4)), np.zeros((4, 4, 4, 4)))
            >>> tree = TernaryTree.from_hamiltonian_coefficients(coeffs)
        """
        if not all([set(coeff.shape) == set(coeffs[0].shape) for coeff in coeffs]):
            logger.error("Coeff axes must be of equal size for all terms.")

        return cls(coeffs[0].shape[0])

    @property
    def enumeration_scheme(self) -> dict[str, tuple[int, int]]:
        """Get the enumeration scheme for the tree.

        Note:
            The tuple is organised as (modes, qubits).

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3).JW()
            >>> tree.enumeration_scheme
            {"": (0,0), "z": (1,1), "zz": (2,2)}
        """
        return self._enumeration_scheme

    @enumeration_scheme.setter
    def enumeration_scheme(self, enumeration_dict: dict[str, tuple[int, int]]):
        """Set the enumeration scheme.

        Args:
            enumeration_dict (dict[str, tuple[int, int]]): An dictionary mapping tree nodes to (mode, qubit) indices
        """
        logger.debug("Setting enumeration scheme.")
        error_string = ""
        if set(self.root_node.child_strings) != set(enumeration_dict.keys()):
            error_string += f"Enumeration scheme {enumeration_dict} must contain all nodes {self.root_node.child_strings}.\n"

        modes = set()
        qubits = set()
        for m, q in enumeration_dict.values():
            logger.debug(f"{m=}{q=}")
            modes.add(m)
            qubits.add(q)
        expected_modes = set(range(self.n_modes))
        if set(modes).symmetric_difference(expected_modes):
            error_string += f"Invalid mode labels {set(modes)} in enumeration scheme ({expected_modes=}).\n"
        if len(set(qubits)) != self.n_modes:
            error_string += f"Expected {self.n_modes} qubit labels, got {len(set(qubits))} in enumeration scheme.\n"

        if error_string != "":
            logger.error(error_string)
            raise ValueError(error_string)

        self.default_mode_op_map = [enum[0] for enum in enumeration_dict.values()]
        self._enumeration_scheme = enumeration_dict

    def encode_topphatt(self, fham:FermionHamiltonian) -> QubitHamiltonian:
        sigs, coeffs = fham.signatures_and_coefficients
        ipow, sym = topphatt(
            flatpack=self.flatpack(),
            n_qubits=self.n_qubits,
            signatures=sigs,
            coeffs=coeffs,
        )

        self._build_symplectic_matrix = lambda : (ipow, sym)
        self.default_mode_op_map = [*range(self.n_modes)]
        return self.encode(fham)

    def flatpack(self) -> list[tuple[int, tuple[int|None, int|None, int|None]]]:
        """Create a TTFlatpack from the tree, which can be passed to rust functions.

        Returns:
            list[tuple[int, tuple[int,int,int]]]
        """
        flatpack: list[tuple[int, tuple[int|None, int|None, int|None]]] = []

        to_flatten: list[TTNode] = [self.root_node]
        while len(to_flatten) > 0:
            node: TTNode = to_flatten.pop(0)
            children: list[int|None] = [None, None,None]
            if isinstance(node.x, TTNode):
                to_flatten.append(node.x)
                children[0] = self.enumeration_scheme[node.x.root_path][1]
            if isinstance(node.y, TTNode):
                to_flatten.append(node.y)
                children[1] = self.enumeration_scheme[node.y.root_path][1]
            if isinstance(node.z, TTNode):
                to_flatten.append(node.z)
                children[2] = self.enumeration_scheme[node.z.root_path][1]

            flatpack.append((int(self.enumeration_scheme[node.root_path][1]), (children[0], children[1], children[2])))

        return flatpack

    def default_enumeration_scheme(self) -> dict[str, tuple[int, int]]:
        """Create a default enumeration scheme for the tree.

        Note:
            The tuple is organised as (modes, qubits).

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3).JW()
            >>> tree.default_enumeration_scheme()
            {"": (0,0), "z": (1,1), "zz": (2,2)}
        """
        logger.debug("Setting default enumeration scheme")
        logger.debug("Child strings %s", self.root_node.child_strings)
        enumeration_scheme = {}
        child_labels = self.root_node.child_qubit_labels
        spare_labels: set[int] = set(range(len(child_labels))).difference(
            child_labels.values()
        )
        for mode, (child, qubit) in enumerate(child_labels.items()):
            if qubit is None:
                qubit = spare_labels.pop()
            enumeration_scheme[child] = (int(mode), int(qubit))
        return enumeration_scheme

    def as_dict(self):
        """Return the tree structure as a dictionary."""
        return self.root_node.as_dict()

    def add_node(self, node_string: str) -> "TernaryTree":
        """Add a node to the tree.

        Args:
            node_string (str): The string representation of the node.

        Returns:
            TernaryTree: The tree with the node added.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
        """
        logger.debug("Adding node %s to TernaryTree", node_string)
        node_string = node_string.lower()
        valid_string = np.all([char in ["x", "y", "z"] for char in node_string])
        if not valid_string:
            raise ValueError("Branch string can only contain x,y,z")

        node = self.root_node
        for char in node_string:
            if isinstance(getattr(node, char), TTNode):
                node = getattr(node, char)
            else:
                node = node.add_child(
                    which_child=char,
                )
        return self

    @property
    def branch_pauli_map(self) -> dict[str, str]:
        """Create a map from each branch string to a Pauli string.

        Returns:
            dict[str, str]: A dictionary of all branch strings with their corresponding Pauli strings.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
            >>> tree.add_node('xz')
            >>> tree.branch_pauli_map
            {'xx': 'XXI',
            'xzx': 'XZX',
            'y': 'YII',
            'xy': 'XYI',
            'xzz': 'XZZ',
            'xzy': 'XZY',
            'z': 'ZII'}
        """
        logger.debug("Building branch operator map for TernaryTree.")

        branches = self.root_node.branch_strings

        qubit_index = {
            node: qubit for node, (_, qubit) in self.enumeration_scheme.items()
        }
        branch_pauli_map = {}
        for branch in branches:
            branch_pauli_map[branch] = ["I"] * self.n_qubits
            node = self.root_node
            for char in branch:
                node_index = qubit_index[node.root_path]
                branch_pauli_map[branch][node_index] = char.upper()
                node = getattr(node, char, None)

            branch_pauli_map[branch] = "".join(branch_pauli_map[branch])
        logger.debug("Branch pauli map complete")
        logger.debug(branch_pauli_map)
        return branch_pauli_map

    @property
    def string_pairs(self) -> dict[str | int, tuple[str, str]]:
        """Return the pair of branch strings which correspond to each node.

        Returns:
            dict[str, tuple(str,str)]: A dictionary of all node labels, j,  with branch strings (2j, 2j+1).

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
            >>> tree.add_node('xz')
            >>> tree.string_pairs
            {'': ('xzz', 'y'), 'x': ('xx', 'xy'), 'xz': ('xzx', 'xzy')}
        """
        logger.debug("Building string pairs for TernaryTree.")
        node_set = self.root_node.child_strings

        pairs = {}
        for node_string in node_set:
            node = self.root_node
            for char in node_string:
                node = getattr(node, char)

            x_string = node_string + "x"
            y_string = node_string + "y"
            while x_string in node_set:
                x_string += "z"

            while y_string in node_set:
                y_string += "z"

            if x_string.count("y") % 2 == 0:
                pairs[node.root_path] = x_string, y_string
            elif y_string.count("y") % 2 == 0:
                pairs[node.root_path] = y_string, x_string

        return pairs

    def _build_symplectic_matrix(
        self,
    ) -> tuple[NDArray[np.uint8], NDArray[bool]]:
        """Build the symplectic matrix for the tree.

        Returns:
            NDArray[np.uint8]: Powers of i for each row of the symplectic matrix.
            NDArray[np.uint8]: Symplectic matrix.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> tree = TernaryTree(3)
            >>> tree.add_node('x')
            >>> tree.enumeration_scheme = tree.default_enumeration_scheme()
            >>> tree._build_symplectic_matrix()
            (array([0, 1, 0, 1, 0, 1], dtype=uint8),
            array([[ True, False, False, False,  True,  True],
                    [ True, False, False,  True, False, False],
                    [ True,  True, False, False, False, False],
                    [ True,  True, False, False,  True, False],
                    [ True, False,  True, False,  True, False],
                    [ True, False,  True, False,  True,  True]]))
        """
        logger.debug("Building symplectic matrix for TernaryTree.")
        if self.enumeration_scheme is None or self.enumeration_scheme == {}:
            logger.warning("No enumeration scheme provided, using default.")
            self.enumeration_scheme = self.default_enumeration_scheme()
        logger.debug(f"{self.enumeration_scheme=}")

        branch_majorana_map = self.root_node.branch_majorana_map
        logger.debug(f"{branch_majorana_map=}")
        if None in [
            *branch_majorana_map.values()
        ] or self.root_node.branch_strings.symmetric_difference(
            branch_majorana_map.keys()
        ):
            logger.info("Branches do not have majorana indices assigned for all modes.")
            logger.info(
                "Using string-pairing algorithm to assign majorana indices to branches."
            )
            branch_majorana_map = string_pairing_algorithm(self)

        pauli_string_map = self.branch_pauli_map

        symplectic = np.zeros((2 * self.n_modes, 2 * self.n_qubits), dtype=bool)
        ipowers = np.zeros((2 * self.n_modes), dtype=np.uint8)
        for operator, majorana_index in branch_majorana_map.items():
            if "x" not in operator and "y" not in operator:
                continue

            operator = pauli_string_map[operator]
            operator = np.array(list(operator), dtype=str)
            # If the string is X or Y then assign 1
            symplectic_term, term_ipower = self._pauli_to_symplectic(operator, 0)

            ipowers[majorana_index] = term_ipower
            symplectic[majorana_index] = symplectic_term
        return ipowers, symplectic

    def JordanWigner(self) -> "TernaryTree":
        """Create a new tree with the Jordan-Wigner encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> jw_tree = TernaryTree(3).JordanWigner()
        """
        return JordanWigner(self.n_modes)

    def JW(self) -> "TernaryTree":
        """Alias for Jordan-Wigner encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> jw_tree = TernaryTree(3).JW()
        """
        return JordanWigner(self.n_modes)

    def ParityEncoding(self) -> "TernaryTree":
        """Create a new tree with the parity encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> parity_tree = TernaryTree(3).ParityEncoding()
        """
        return ParityEncoding(self.n_modes)

    def BravyiKitaev(self) -> "TernaryTree":
        """Create a new tree with the Bravyi-Kitaev encoding.

        Args:
            n_modes (int): The number of fermionic modes.

        Returns:
            TernaryTree: A ternary tree encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> bk_tree = TernaryTree(3).BravyiKitaev()
        """
        return BravyiKitaev(self.n_modes)

    def BK(self) -> "TernaryTree":
        """Alias for Bravyi-Kitaev encoding.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> bk_tree = TernaryTree(3).BK()
        """
        return BravyiKitaev(self.n_modes)

    def JKMN(self) -> "TernaryTree":
        """Create a new tree with the JKMN encoding.

        The JKMN encoding gives a ternary tree with the minimum Pauli-weight.

        Example:
            >>> from ferrmion.encode.ternary_tree import TernaryTree
            >>> min_height_tree = TernaryTree(3).JKMN()
        """
        return JKMN(self.n_modes)

def string_pairing_algorithm(tree: TernaryTree):
    """String-pairing algoritm.

    This is used to produce a map from branches to majorana-indices
    of the root node.

    Args:
        tree (TernaryTree): A Ternary-tree encoding.

    Returns:
        dict[str, int]: A map from branches to majorana mdoe indices.
    """
    logger.debug("Running the string-pairing algorithm.")
    node_set = tree.root_node.child_strings

    branch_majorana_map = {}
    for node_string in node_set:
        # We want to set the majorana indices according to the
        # fermionic ones so that f_i -> (m_2i, m_2i+1)
        fermion_mode = tree.enumeration_scheme[node_string][0]

        x_string = node_string + "x"
        y_string = node_string + "y"
        while x_string in node_set:
            x_string += "z"

        while y_string in node_set:
            y_string += "z"

        if x_string.count("y") % 2 == 0:
            branch_majorana_map[x_string] = 2 * fermion_mode
            branch_majorana_map[y_string] = 2 * fermion_mode + 1
        elif y_string.count("y") % 2 == 0:
            branch_majorana_map[y_string] = 2 * fermion_mode
            branch_majorana_map[x_string] = 2 * fermion_mode + 1

    # We'll place the all-z string after all the required majorana modes
    all_z = "z"
    while all_z in node_set:
        all_z += "z"
    branch_majorana_map[all_z] = 2 * len(node_set) + 1

    logger.debug("String-paring algoithm complete.")
    logger.debug(f"{branch_majorana_map=}")
    return branch_majorana_map


def JordanWigner(n_modes: int) -> TernaryTree:
    """Create a new tree with the Jordan-Wigner encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import JordanWigner
        >>> jw_tree = JordanWigner(3)
    """
    logger.debug("Creating Jordan-Wigner encoding tree")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    new_tree.add_node("z" * (n_modes - 1))
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree


def JW(n_modes: int) -> TernaryTree:
    """Alias for Jordan-Wigner encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import JW
        >>> jw_tree = JW(3)
    """
    return JordanWigner(n_modes)


def ParityEncoding(n_modes: int) -> TernaryTree:
    """Create a new tree with the parity encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import ParityEncoding
        >>> parity_tree = ParityEncoding(3)
    """
    logger.debug("Creating parity encoding tree")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    new_tree.add_node("x" * (n_modes - 1))
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree


def PE(n_modes: int) -> TernaryTree:
    """Alias for ParityEncoding"""
    return ParityEncoding(n_modes)


def BravyiKitaev(n_modes: int) -> "TernaryTree":
    """Create a new tree with the Bravyi-Kitaev encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import BravyiKitaev
        >>> bk_tree = BravyiKitaev(3)
    """
    logger.debug("Creating Bravyi-Kitaev encoding tree")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    branches = ["x"]
    # one is used for root, which is defined
    remaining_qubits = n_modes - 1
    while remaining_qubits > 0:
        new_branches = set()
        for item in branches:
            if remaining_qubits > 0:
                new_tree.add_node(item)
                remaining_qubits -= 1
            else:
                break

            new_branches.add(item + "x")
            new_branches.add(item + "z")
        branches = sorted(list(new_branches), key=node_sorter)
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree


def BK(n_modes: int) -> TernaryTree:
    """Alias for Bravyi-Kitaev encoding.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import BK
        >>> bk_tree = BK(3)
    """
    return BravyiKitaev(n_modes)


def JKMN(n_modes: int) -> TernaryTree:
    """Create a new tree with the JKMN encoding.

    The JKMN encoding gives a ternary tree with the minimum Pauli-weight.

    Args:
        n_modes (int): The number of fermionic modes.

    Returns:
        TernaryTree: A ternary tree encoding.

    Example:
        >>> from ferrmion.encode.ternary_tree import JKMN
        >>> min_height_tree = JKMN(3)
    """
    logger.debug("Creating JKMN encoding tree.")
    new_tree = TernaryTree(
        n_modes=n_modes,
        root_node=TTNode(),
    )
    branches = ["x", "y", "z"]
    # one is used for root which is defined
    remaining_qubits = n_modes - 1
    while remaining_qubits > 0:
        new_branches = set()
        for item in branches:
            if remaining_qubits > 0:
                new_tree.add_node(item)
                remaining_qubits -= 1
            else:
                break

            new_branches.add(item + "x")
            new_branches.add(item + "y")
            new_branches.add(item + "z")
        branches = sorted(list(new_branches), key=node_sorter)
    new_tree.enumeration_scheme = new_tree.default_enumeration_scheme()
    return new_tree



"""Tests for functions in the optimize submodule."""

from ferrmion.optimize.bonsai import bonsai_algorithm
import rustworkx as rx
import numpy as np
from pytest import fixture

@fixture
def heavy_hex_graph():
    graph = rx.PyGraph()
    graph.add_nodes_from(range(37))
    graph.add_edges_from_no_data(
        [
            (0, 1),
            (0, 2),
            (0, 3),
            (1, 4),
            (2, 5),
            (3, 6),
            (4, 7),
            (4, 8),
            (5, 9),
            (5, 10),
            (6, 11),
            (6, 12),
            (7, 13),
            (8, 14),
            (9, 15),
            (10, 16),
            (11, 17),
            (12, 18),
            (13, 19),
            (13, 20),
            (14, 21),
            (14, 22),
            (15, 23),
            (15, 24),
            (16, 25),
            (16, 26),
            (17, 27),
            (17, 28),
            (18, 29),
            (18, 30),
            (22, 31),
            (26, 32),
            (30, 33),
            (31, 34),
            (32, 35),
            (33, 36),
        ]
    )
    return graph

@fixture
def garnet_graph():
    garnet_edges = [
        (1, 0),
        (1, 4),
        (3, 0),
        (3, 2),
        (3, 4),
        (3, 8),
        (5, 4),
        (5, 6),
        (5, 10),
        (7, 2),
        (7, 8),
        (7, 12),
        (9, 4),
        (9, 8),
        (9, 10),
        (9, 14),
        (11, 6),
        (11, 10),
        (11, 16),
        (13, 8),
        (13, 12),
        (13, 14),
        (13, 17),
        (15, 10),
        (15, 14),
        (15, 16),
        (15, 19),
        (18, 14),
        (18, 17),
        (18, 19)
    ]

    graph = rx.PyGraph()
    graph.add_nodes_from(range(20))
    graph.add_edges_from_no_data(garnet_edges)
    return graph

@fixture
def emerald_graph():
    emerald_edges = [(0, 1),
        (0, 4),
        (3, 2),
        (3, 4),
        (3, 9),
        (5, 1),
        (5, 4),
        (5, 6),
        (5, 11),
        (8, 2),
        (8, 7),
        (8, 9),
        (8, 16),
        (10, 4),
        (10, 9),
        (10, 11),
        (10, 18),
        (12, 11),
        (12, 13),
        (12, 20),
        (15, 7),
        (15, 14),
        (15, 16),
        (15, 23),
        (17, 16),
        (17, 18),
        (17, 25),
        (18, 26),
        (19, 11),
        (19, 18),
        (19, 20),
        (19, 27),
        (21, 13),
        (21, 20),
        (21, 29),
        (22, 14),
        (22, 23),
        (24, 16),
        (24, 23),
        (24, 25),
        (24, 32),
        (25, 26),
        (26, 27),
        (26, 34),
        (28, 20),
        (28, 27),
        (28, 29),
        (28, 36),
        (30, 29),
        (30, 38),
        (31, 23),
        (31, 32),
        (33, 25),
        (33, 32),
        (33, 34),
        (33, 41),
        (34, 42),
        (35, 27),
        (35, 34),
        (35, 36),
        (35, 43),
        (36, 37),
        (37, 38),
        (37, 45),
        (40, 32),
        (40, 39),
        (40, 41),
        (40, 46),
        (41, 42),
        (43, 49),
        (44, 36),
        (44, 43),
        (44, 45),
        (44, 50),
        (47, 41),
        (47, 46),
        (47, 48),
        (47, 51),
        (49, 50),
        (49, 53),
        (52, 48),
        (52, 51),
        (52, 53)
    ]

    graph = rx.PyGraph()
    graph.add_nodes_from(range(54))
    graph.add_edges_from_no_data(emerald_edges)
    return graph

def test_heavy_hex(heavy_hex_graph):
    graph = heavy_hex_graph
    bonsai_homo = bonsai_algorithm(graph=graph, homogenous=True)
    assert bonsai_homo.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xx',
        'yx',
        'zx',
        'xxx',
        'xxy',
        'yxx',
        'yxy',
        'zxx',
        'zxy',
        'xxxx',
        'xxyx',
        'yxxx',
        'yxyx',
        'zxxx',
        'zxyx',
        'xxxxx',
        'xxxxy',
        'xxyxx',
        'xxyxy',
        'yxxxx',
        'yxxxy',
        'yxyxx',
        'yxyxy',
        'zxxxx',
        'zxxxy',
        'zxyxx',
        'zxyxy',
        'xxyxyx',
        'yxyxyx',
        'zxyxyx',
        'xxyxyxx',
        'yxyxyxx',
        'zxyxyxx']

    bonsai_hetero = bonsai_algorithm(graph=graph, homogenous=False)
    assert bonsai_hetero.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xz',
        'yz',
        'zz',
        'xzx',
        'xzz',
        'yzx',
        'yzz',
        'zzx',
        'zzz',
        'xzxz',
        'xzzz',
        'yzxz',
        'yzzz',
        'zzxz',
        'zzzz',
        'xzxzx',
        'xzxzz',
        'xzzzx',
        'xzzzz',
        'yzxzx',
        'yzxzz',
        'yzzzx',
        'yzzzz',
        'zzxzx',
        'zzxzz',
        'zzzzx',
        'zzzzz',
        'xzxzxz',
        'yzxzxz',
        'zzxzxz',
        'xzxzxzz',
        'yzxzxzz',
        'zzxzxzz']


    assert bonsai_hetero.root_node.child_qubit_labels == {
        '': 0,
        'x': 2,
        'y': 3,
        'z': 1,
        'xz': 5,
        'yz': 6,
        'zz': 4,
        'xzx': 10,
        'xzz': 9,
        'yzx': 12,
        'yzz': 11,
        'zzx': 8,
        'zzz': 7,
        'xzxz': 16,
        'xzzz': 15,
        'yzxz': 18,
        'yzzz': 17,
        'zzxz': 14,
        'zzzz': 13,
        'xzxzx': 26,
        'xzxzz': 25,
        'xzzzx': 24,
        'xzzzz': 23,
        'yzxzx': 30,
        'yzxzz': 29,
        'yzzzx': 28,
        'yzzzz': 27,
        'zzxzx': 22,
        'zzxzz': 21,
        'zzzzx': 20,
        'zzzzz': 19,
        'xzxzxz': 32,
        'yzxzxz': 33,
        'zzxzxz': 31,
        'xzxzxzz': 35,
        'yzxzxzz': 36,
        'zzxzxzz': 34
    }

def test_heavy_hex_max_nodes(heavy_hex_graph):
    bonsai= bonsai_algorithm(graph=heavy_hex_graph, homogenous=False, max_nodes=14)
    bonsai.string_pairs
    assert bonsai.as_dict() == {'x': {'x': None,
  'y': None,
  'z': {'x': {'x': None, 'y': None, 'z': None},
   'y': None,
   'z': {'x': None, 'y': None, 'z': None}}},
 'y': {'x': None,
  'y': None,
  'z': {'x': {'x': None, 'y': None, 'z': None},
   'y': None,
   'z': {'x': None, 'y': None, 'z': None}}},
 'z': {'x': None,
  'y': None,
  'z': {'x': {'x': None, 'y': None, 'z': None},
   'y': None,
   'z': {'x': None, 'y': None, 'z': {'x': None, 'y': None, 'z': None}}}}}

def test_garnet(garnet_graph):
    graph = garnet_graph
    bonsai= bonsai_algorithm(graph=graph, homogenous=False)
    assert bonsai.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xx',
        'xz',
        'yx',
        'yz',
        'zx',
        'zy',
        'zz',
        'xxx',
        'xxz',
        'xzz',
        'yxz',
        'yzz',
        'zxz',
        'zyz',
        'zzz',
        'xxzz']

    bonsai= bonsai_algorithm(graph=graph, homogenous=True)
    assert bonsai.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xx',
        'xy',
        'xz',
        'yx',
        'yy',
        'zx',
        'zy',
        'xxx',
        'xyx',
        'xzx',
        'yxx',
        'yyx',
        'yyy',
        'zxx',
        'zyx',
        'yyxx']


def test_emerald_hom(emerald_graph):
    graph = emerald_graph.copy()
    bonsai= bonsai_algorithm(graph=graph, homogenous=False)
    assert bonsai.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xx',
        'xz',
        'yx',
        'yz',
        'zx',
        'zz',
        'xxx',
        'xxz',
        'xzx',
        'xzz',
        'yxz',
        'yzx',
        'yzz',
        'zxx',
        'zxz',
        'zzx',
        'zzz',
        'xxxx',
        'xxxz',
        'xxzz',
        'yzxz',
        'yzzx',
        'yzzz',
        'zxxz',
        'zxzx',
        'zxzz',
        'zzxz',
        'zzzx',
        'zzzy',
        'zzzz',
        'xxzzz',
        'yzxzz',
        'yzzxx',
        'yzzxz',
        'yzzzz',
        'zxxzz',
        'zxzxz',
        'zxzzx',
        'zxzzz',
        'zzzzz',
        'xxzzzx',
        'xxzzzz',
        'yzxzzz',
        'yzzxxz',
        'yzzxzz',
        'yzzzzz',
        'zxzxzz',
        'zxzzxz',
        'zxzzzz',
        'xxzzzzz']


def test_emerald_het(emerald_graph):
    graph = emerald_graph.copy()
    bonsai= bonsai_algorithm(graph=graph, homogenous=True)
    assert bonsai.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xx',
        'xy',
        'yx',
        'yy',
        'zx',
        'zy',
        'xxx',
        'xxy',
        'xyx',
        'xyy',
        'yxx',
        'yxy',
        'yyx',
        'yyy',
        'zxx',
        'zxy',
        'zyx',
        'xxxx',
        'xxxy',
        'xxxz',
        'xxyx',
        'xyxx',
        'xyxy',
        'xyyx',
        'yyxx',
        'yyyx',
        'yyyy',
        'zxxx',
        'zxxy',
        'zxyx',
        'xxxxx',
        'xyxxx',
        'xyxxy',
        'xyxyx',
        'xyyxx',
        'yyxxx',
        'zxxxx',
        'zxxyx',
        'zxxyy',
        'zxyxx',
        'xyxxxx',
        'xyxxyx',
        'xyxyxx',
        'yyxxxx',
        'yyxxxy',
        'zxxxxx',
        'zxxyxx',
        'zxxyyx',
        'zxyxxx',
        'yyxxxxx']

def test_emerald_het_max_nodes(emerald_graph):
    graph = emerald_graph
    bonsai = bonsai_algorithm(graph=graph, homogenous=False, max_nodes=14)
    assert bonsai.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xx',
        'xz',
        'yx',
        'yz',
        'zx',
        'zz',
        'zxx',
        'zxz',
        'zzx',
        'zzz']

def test_emerald_hom_max_nodes(emerald_graph):
    graph = emerald_graph
    bonsai = bonsai_algorithm(graph=graph, homogenous=True, max_nodes=14)
    assert bonsai.root_node.child_strings == [
        '',
        'x',
        'y',
        'z',
        'xx',
        'xy',
        'yx',
        'yy',
        'zx',
        'zy',
        'xxx',
        'xxy',
        'xyx',
        'xyy']



"""Tests for Utils functions"""

import numpy as np
from ferrmion.utils import (
    icount_to_sign,
    pauli_to_symplectic,
    symplectic_hash,
    symplectic_to_pauli,
    symplectic_to_sparse,
    symplectic_unhash,
    fermionic_to_sparse_majorana,
)


def test_icount_to_sign() -> None:
    assert icount_to_sign(0) == 1
    assert icount_to_sign(1) == 1j
    assert icount_to_sign(2) == -1
    assert icount_to_sign(3) == -1j


def test_symplectic_hashing() -> None:
    symplectic = np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)
    print(symplectic_hash(symplectic))
    print(symplectic_unhash(symplectic_hash(symplectic), len(symplectic)))


def test_symplectic_pauli_conversion() -> None:
    symplectic = np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)

    assert symplectic_to_pauli(symplectic, 0) == ("IZXY", 3)
    assert symplectic_to_pauli(symplectic) == ("IZXY", 3)
    inverse_symplectic, inverse_ipower = pauli_to_symplectic(
        *symplectic_to_pauli(symplectic, 0)
    )
    assert inverse_ipower == 0
    assert np.all(inverse_symplectic == symplectic)


def test_symplectic_sparse_conversion() -> None:
    symplectic = np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)

    assert symplectic_to_sparse(symplectic, 1)[0] == "ZXY"
    assert symplectic_to_sparse(symplectic, 1)[2] == 1.0
    assert np.array_equal(symplectic_to_sparse(symplectic, 1)[1], [1, 2, 3])


def test_fermionic_to_sparse_majorana() -> None:
    n_modes = 3
    ones = np.zeros((n_modes, n_modes))
    twos = np.zeros((n_modes, n_modes, n_modes, n_modes))

    ones[0, 0] = 1
    twos[1, 2, 1, 2] = 2

    majorana_ham = fermionic_to_sparse_majorana([(ones, "+-"), (twos, "++--")])
    assert majorana_ham == {
        (0, 1): np.complex128(0.5j),
        (4, 5): np.complex128(-0.5j),
        (2, 3): np.complex128(-0.5j),
        (2, 3, 4, 5): np.complex128(0.5 + 0j),
    }



"""Init for fermion qubit encodings.

This file is ignored by pre-commit as the pyo3 integration requires importing
rust functions before importing functions from the python module.
"""

from .core import hartree_fock_state, symplectic_product
from .encode import FermionQubitEncoding 
from .encode.maxnto import MaxNTO, maxnto_symplectic_matrix
from .encode.ternary_tree import TernaryTree
from .encode.ternary_tree_node import TTNode, node_sorter
from .encode.standard import (
    jordan_wigner,
    jordan_wigner_annealed,
    jordan_wigner_topphatt,
    parity,
    parity_annealed,
    parity_topphatt,
    bravyi_kitaev,
    bravyi_kitaev_annealed,
    bravyi_kitaev_topphatt,
    jkmn,
    jkmn_annealed,
    jkmn_topphatt,
)
from .hamiltonians import FermionHamiltonian, molecular_hamiltonian, hubbard_hamiltonian
from .utils import (
    icount_to_sign,
    pauli_to_symplectic,
    setup_logs,
    symplectic_hash,
    symplectic_to_pauli,
    symplectic_unhash,
    two_operator_product,
    
)

__all__ = [
    "FermionQubitEncoding",
    "TernaryTree",
    "TTNode",
    "node_sorter",
    "pauli_to_symplectic",
    "symplectic_to_pauli",
    "symplectic_hash",
    "symplectic_unhash",
    "symplectic_product",
    "icount_to_sign",
    "MaxNTO",
    "maxnto_symplectic_matrix",
    "hartree_fock_state",
    "two_operator_product",
    "FermionHamiltonian",
    "molecular_hamiltonian",
    "hubbard_hamiltonian",
    "FermionQubitEncoding",
    "TernaryTree",
    "MaxNTO",
    "jordan_wigner",
    "jordan_wigner_annealed",
    "jordan_wigner_topphatt",
    "parity",
    "parity_annealed",
    "parity_topphatt",
    "bravyi_kitaev",
    "bravyi_kitaev_annealed",
    "bravyi_kitaev_topphatt",
    "jkmn",
    "jkmn_annealed",
    "jkmn_topphatt",
]

setup_logs()
